# -*- coding: utf-8 -*-
"""
gstorage.utils
~~~~~~~~~~~~~~

Helper functions used by other modules
"""
import os

from django.conf import settings


def get_config(key):
    """
    Get the value of config from settings. Else get it from environment.

    :type key: string
    :param key: A string indicating the key to config, for ex - GOOGLE_APPLICATION_CREDENTIALS

    :returns: The value of the key if present, else None
    """
    try:
        return getattr(settings, key)
    except AttributeError:
        try:
            return os.environ[key]
        except KeyError:
            pass
    return None


def find_files(path):
    """
    Return only the files under the given path

    :type key: string
    :param path: relative or absolute path to the directory
    :return: List of string where each item is a filepath
    """
    return [os.path.join(dirpath, file)
            for (dirpath, dirs, files) in os.walk(path)
            for file in files]



# -*- coding: utf-8 -*-
"""
gstorage.bucket
~~~~~~~~~~~~~~~
A thin wrapper around gcloud.bucket with convenience
methods to interact with buckets

    >>> from gstorage.bucket import Bucket
    >>> bucket = Bucket.get_default()
    >>> bucket.copydir('/share/images')
"""
import os

from gcloud.storage.blob import Blob
from gcloud.storage.bucket import Bucket as BaseBucket
from gcloud.storage.client import Client
from oauth2client.client import GoogleCredentials

from .utils import find_files, get_config


class Bucket(BaseBucket):

    @classmethod
    def get_or_create(cls, bucket_name, client=None):
        """
        If the bucket exists with this name, get it. Else, create it

        :param cls: :class:`gstorage.bucket.Bucket`
        :type bucket_name: string
        :param bucket_name: name of the bucket
        :type client: gcloud.client.Client
        :param client: (optional) instance of client to use
        :return: :class:`Bucket <Bucket>` object
        :raises gcloud.exceptions.BadRequest (400): not a valid bucket name
        :raises gcloud.exceptions.Forbidden (403): The credentials are invalid
        """
        if not client:
            credentials = GoogleCredentials.get_application_default()
            client = Client(credentials=credentials)
        bucket = cls(client, name=bucket_name)
        if not bucket.exists():
            bucket.create()
        return bucket

    def copydir(self, path):
        """
        Copy the contents of the local directory given by path to google cloud.
        Maintain the same directory structure on remote.
        This is (intentionally) a blocking call, so clients can report errors if
        the transfer fails.

        :type path: string
        :param path: relative or absolute path to the directory that needs to be copied
        :return: True when transfer is complete
        :raises OSError: path doesn't exist or permission denied
        :raises ValueError: if the library cannot determine the file size
        :raises gcloud.exceptions.GCloudError: if upload status gives error response
        """
        if not os.access(path, os.R_OK):
            raise OSError('Permission denied')
        for filename in find_files(path):
            blob = Blob(filename, self)
            blob.upload_from_filename(filename)
        return True

    @classmethod
    def get_default(cls):
        """
        Get an instance of the default bucket identified fy GCLOUD_DEFAULT_BUCKET_NAME

        :param cls: :class:`gstorage.bucket.Bucket`
        :return: :class:`Bucket <Bucket>` object
        :raises gcloud.exceptions.BadRequest (400): not a valid bucket name
        :raises gcloud.exceptions.Forbidden (403): The credentials are invalid
        """
        return cls.get_or_create(get_config('GCLOUD_DEFAULT_BUCKET_NAME'))



# -*- coding: utf-8 -*-
"""
Check that the module is registered with required settings
"""
from django.core import checks

from .constants import REQUIRED_SETTINGS
from .utils import get_config


def check_gstorage_params(**kwargs):
    """
    Check that the module is initialized with all the required settings/environment

    Args:
        kwargs: optional list of app_config

    Returns:
        A list of errors
    """
    errors = []
    module_name = 'gstorage'

    for index, field in enumerate(REQUIRED_SETTINGS):
        if not get_config(field):
            errors.append(
                checks.Error(
                    '%s is required' % field,
                    hint=None,
                    obj=module_name,
                    id='%s.00%s' % (module_name, index + 1)  # use 1-based index for errors
                )
            )
    return errors



# -*- coding: utf-8 -*-
"""
Specify the config for the module used to install the app
"""
from django.apps import AppConfig
from django.core import checks
from django.utils.translation import ugettext_lazy as _

from .checks import check_gstorage_params


class GStorageTestConfig(AppConfig):
    """A test config which initializes the app without checks at startup"""
    name = 'gstorage'
    verbose_name = _('Django GStorage')


class GStorageConfig(GStorageTestConfig):

    def ready(self):
        checks.register(checks.Tags.models)(check_gstorage_params)



# -*- coding: utf-8 -*-
"""
gstorage.storage
~~~~~~~~~~~~~~~~

Implement the interface expected by :class:`django.core.files.storage.Storage`

    >>> from django.core.files.uploadfile import TemporaryUploadedFile
    >>> from gstorage.storage import Storage
    >>> storage = Storage(location='media/2016')
    >>> with TemporaryUploadedFile('test', 'text/plain', 1, 'utf8') as content:
    ...     storage.save('test.txt', content)
    ...
    >>> u'media/2016/test'
"""
import os

from django.core.files.storage import Storage as BaseStorage
from django.utils.crypto import get_random_string
from gcloud.storage.blob import Blob

from .bucket import Bucket


class Storage(BaseStorage):

    def __init__(self, location=None, base_url=None, bucket=None):
        self._location = location
        self._base_url = base_url
        if not bucket:
            self._bucket = Bucket.get_default()
        super(Storage, self).__init__()

    def _open(self, name, mode):
        """
        Retrieves the specified file from storage.
        """
        return Blob(name, self._bucket)

    def _save(self, name, content):
        """
        Saves new content to the file specified by name. The content should be
        a proper File object or any python file-like object, ready to be read
        from the beginning.
        """
        path = os.path.join(self._location, content.name) if self._location else content.name
        blob = Blob(path, self._bucket)
        blob.upload_from_file(content, size=content.size)
        return blob.name

    def get_valid_name(self, name):
        """
        Returns a filename, based on the provided filename, that's suitable for
        use in the target storage system.
        """
        return name

    def get_available_name(self, name):
        """
        Returns a filename that's free on the target storage system, and
        available for new content to be written to.
        """
        dir_name, file_name = os.path.split(name)
        file_root, file_ext = os.path.splitext(file_name)
        # If the filename already exists, add an underscore and a random 7
        # character alphanumeric string (before the file extension, if one
        # exists) to the filename until the generated filename doesn't exist.
        while self.exists(name):
            # file_ext includes the dot.
            name = os.path.join(dir_name, "%s_%s%s" % (file_root, get_random_string(7), file_ext))
        return name

    def path(self, name):
        """
        Returns a local filesystem path where the file can be retrieved using
        Python's built-in open() function. Storage systems that can't be
        accessed using open() should *not* implement this method.
        """
        raise NotImplementedError("This backend doesn't support absolute paths.")

    def delete(self, name):
        """
        Deletes the specified file from the storage system.
        """
        raise NotImplementedError('subclasses of Storage must provide a delete() method')

    def exists(self, name):
        """
        Returns True if a file referenced by the given name already exists in the
        storage system, or False if the name is available for a new file.
        """
        return Blob(name, self._bucket).exists()

    def listdir(self, path):
        """
        Lists the contents of the specified path, returning a 2-tuple of lists;
        the first item being directories, the second item being files.
        """
        raise NotImplementedError('subclasses of Storage must provide a listdir() method')

    def size(self, name):
        """
        Returns the total size, in bytes, of the file specified by name.
        """
        raise NotImplementedError('subclasses of Storage must provide a size() method')

    def url(self, name):
        """
        Returns an absolute URL where the file's contents can be accessed
        directly by a Web browser.
        """
        return Blob(name, self._bucket).public_url

    def accessed_time(self, name):
        """
        Returns the last accessed time (as datetime object) of the file
        specified by name.
        """
        raise NotImplementedError('subclasses of Storage must provide an accessed_time() method')

    def created_time(self, name):
        """
        Returns the creation time (as datetime object) of the file
        specified by name.
        """
        raise NotImplementedError('subclasses of Storage must provide a created_time() method')

    def modified_time(self, name):
        """
        Returns the last modified time (as datetime object) of the file
        specified by name.
        """
        raise NotImplementedError('subclasses of Storage must provide a modified_time() method')



"""
    django-gstorage
    ~~~~~~~~~~~~~~~
"""
import re

from setuptools import setup

version = ''
with open('gstorage/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        fd.read(), re.MULTILINE).group(1)

with open('README.rst', 'r') as f:
    readme = f.read()
with open('HISTORY.rst', 'r') as f:
    history = f.read()

setup(
    name='django-gstorage',
    version=version,
    description='Allow easy integration with Google storage for Django projects',
    long_description=readme + '\n\n' + history,
    author='Pradip Caulagi',
    author_email='caulagi@gmail.com',
    url='http://github.com/fyndiq/django-gstorage/',
    packages=['gstorage'],
    package_data={'': ['LICENSE', 'HISTORY.rst']},
    include_package_data=True,
    install_requires=[
        'Django',
        'gcloud',
    ],
    license='MIT',
    zip_safe=False,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



# -*- coding: utf-8 -*-
"""
    django-gstorage
    ~~~~~~~~~~~~~~~

    A plug-and-play replacement for FileSystemStorage but
    using Google storage for persistence
"""

__version__ = '0.5.1'
__author__ = 'Fyndiq'
__license__ = 'MIT'
__copyright__ = 'Copyright 2016 Fyndiq AB'


default_app_config = 'gstorage.apps.GStorageConfig'



# -*- coding: utf-8 -*-
"""
Collect all the constants required by the module in one place
"""

# Settings that are required for the library to perform correctly
REQUIRED_SETTINGS = (
    'GOOGLE_APPLICATION_CREDENTIALS',
    'GCLOUD_PROJECT_NAME',
    # default bucket that should be used to store assets
    'GCLOUD_DEFAULT_BUCKET_NAME',
)
